/**
* Loads functionality to compose dynamic HTML.
*/
HTMLComposer = function() {
}

HTMLComposer._cache = new Object();
HTMLComposer._idcache = new Object();

/**
 * Returns a DigComposer instance for the specified text.  If a
 * composer already exists, return a "copy" of the cached
 * HTMLComposer.
 *
 * @param  text (String) The HTML text for this HTMLComposer.
 * @oparam id (String) The returned instance is accessible as a property of
 *         the composer node using the specified
 *         <code>id</code>.  For example:
 * <plaintext>
 * HTMLComposer.create('<h3>@Name@ Player </h2>', 'treeView');
 * HTMLComposer.treeView.replaceToken('Name', 'Radio');
 * HTMLComposer.treeView.compose('divElementID');
 * </plaintext>
 *
 * @return (DigComposer)
 * @factory
 */
HTMLComposer.create = function(text, id) {
    // Parse the text into a composer if its not in the cache
    var composer = HTMLComposer._cache[id];
    if (!composer) {
        composer = new DigComposer(text);
        HTMLComposer._cache[id] = composer;
    }

    // Return a unique instance to allow different replacements
    // for the same text
    var retval = composer._clone();


    return retval;
}

/**
 * Get an array of all the keys in <code>anObject</code>.
 *
 * @param anObject The object that contains the keys.
 * @return (Array) An array of all the object's keys.
 */
HTMLComposer.keys = function(anObject) {
    var myList = new Array();
    for (var myKey in anObject) {
        myList[myList.length] = myKey;
    }

    return myList;
}
/**
 * Clear all the values of <code>anObject</code> by setting the value of
 * each key to null.
 *
 * @param anObject The object that will have its values cleared.
 */
HTMLComposer.clearValues = function(anObject) {

    for (var myKey in anObject) anObject[myKey] = null;
}

/**
 * Test <code>if typeof(aThing) == 'string'</code>.
 *
 * @param aThing The value to test.
 * @return (Boolean)
 */
HTMLComposer.isString = function(aThing) {
    return (typeof(aThing) == 'string');
}
/**
 * Test <code>if (aThing === void 0)</code>.
 *
 * @param aThing The value to test.
 * @return (Boolean)
 */
HTMLComposer.isUndefined = function(aThing) {
    return (aThing === void 0);
}

/**
 * Test <code>if (aThing !== void 0) && (aThing !== null)</code>.
 * @param aThing The value to test.
 * @return (Boolean)
 */
HTMLComposer.isSpecified = function(aThing) {
    return ((aThing !== void 0) && (aThing !== null));
}

/**
 * Test <code> if aThing</code> is an instance of <code>aClass</code>.
 *
 * @param aThing The instance to check.
 * @param aClass The class of the object
 * @return (Boolean)
 */
HTMLComposer.isInstance = function(aThing, aClass) {
    return !!(aThing && aThing.constructor && (aThing.constructor === aClass));
}

// -----------------------------------------------------------
// Internal Classes
// -----------------------------------------------------------

/**
 * This class simply stores the name of the token.
 *
 * @class _Token
 */

/**
 * Create a Token chunk.  This class simply stores the name of the token.
 *
 * @param name (String) Name of the token (e.g. "lastName")
 * @param nameWithDelimeter (String) Name with the token delimeters (@foo@)
 * @constructor
 */
_Token = function(name, nameWithDelimeter) {
    this._name = name;
    this._nameWithDelimeter = nameWithDelimeter;
}
// -----------------------------------------------------------
// DigComposer class
// -----------------------------------------------------------

/**
 * <p>This class provides an API to replace tokens and token
 * blocks included as part of an HTML string as described in the
 * <code>composer</code> documentation.  Once all replacements
 * have been made, the contents are composed into a string.  The
 * <code>HTMLComposer.create()</code> factory method returns
 * instances of this class.</p>
 *
 * <h4>Examples</h4>
 *
 * <p class=fwQuestion> 
 * &rarr; Replacing Tokens and Removing Token Blocks</p>
 *
 * <blockquote>
 * <plaintext>
 * var str = "
 *     Welcome @name@!.
 *     <!-- @EmployeeBlock@ -->
 *     Mr. @name@, the AOL stock is trading at $50.
 *     <!-- @/EmployeeBlock@ -->."
 *
 * // Create the composer for the HTML string.
 * var pageComposer = HTMLComposer.create(str);
 * // Replaces *all* occurences of the name token in the HTML
 * // including the token inside of the Employee Block
 * // (assuming it doesn't get removed).
 * pageHTMLComposer.replaceToken('name', 'John Doe');
 * if (!user.isEmployee()) {
 *     // Replacing a token/token block with null removes it.
 *     pageHTMLComposer.replaceTokenBlock('EmployeeBlock', null);
 * }
 * var generatedContent = pageHTMLComposer.compose();
 * </plaintext>
 * </blockquote>
 *
 * <p class=fwQuestion> &rarr; Conditional Removal of an HTML Block</p>
 *
 * As shown in the above example, token blocks can be used to
 * conditionally remove a block of HTML.
 *
 * <p class=fwQuestion> &rarr; Looping</p>
 *
 * <p>Looping over a block of HTML is useful when you are dynamically
 * generating rows of a table, items in a list, etc.
 * Token blocks can be used to accomplish looping.
 * Here is an example that dynamically generates the rows of a
 * table:</p>
 *
 * <blockquote>
 * <plaintext>
 * var str = "
 *     <table>
 *     <!-- @RowBlock@ -->
 *     <tr><td>@date@</td><td>@description@</td></tr>
 *     <!-- @/RowBlock@ -->
 *     </table>";
 *
 * // Create the composer for the HTML string and one for the
 * // block we want to iterate over.
 * var pageComposer = HTMLComposer.create(str);
 * var rowComposer = pageHTMLComposer.getTokenBlock('RowBlock');
 *
 * for (int i = 0; i < items.length; i++) {
 *     // On each iteration, replace the tokens in the block and
       // call "next".  This stores the contents of this iteration
 *     rowHTMLComposer.replaceToken('date', items[i].date);
 *     rowHTMLComposer.replaceToken('description', items[i].description);
 *     rowHTMLComposer.next();
 * }
 * 
 * // Compose the contents into a DIV
 * pageHTMLComposer.compose('myDiv');
 * </plaintext>
 * </blockquote>
 *
 * <p class=fwQuestion> &rarr; Nested Token Blocks </p>
 *
 * <p>The composer supports infinite level of token block nesting.
 * *All* tokens and token blocks are visible to the top level
 * composer including nested tokens and token blocks as shown in
 * the example below:</p>
 *
 * <blockquote>
 * <plaintext>
 * var str = "
 *     Welcome @name@!.
 *     <!-- @EmployeeBlock@ -->
 *     Mr. @name@, the AOL stock is trading at $50.
 *         <!-- @EngineerBlock@ -->
 *            Download the @Tool@ from the JavaSoft website!
 *         <!-- @/EngineerBlock@ -->
 *     <!-- @/EmployeeBlock@ -->
 *
 * // Create the composer for the HTML string.
 * var pageComposer = HTMLComposer.create(str);
 * pageHTMLComposer.replaceToken('name', 'John Doe');
 * if (user.isEmployee()) {
 *    if (user.isEngineer()) {
 *        pageHTMLComposer.replaceToken('Tool', 'JDK');
 *    } else {
 *        pageHTMLComposer.replaceTokenBlock('EngineerBlock', null);
 *    }
 * } else {
 *    pageHTMLComposer.replaceTokenBlock('EmployeeBlock', null);
 * }
 * var generatedContent = pageHTMLComposer.compose();
 * </plaintext>
 * </blockquote>
 *
 * @class DigComposer
 * @see HTMLComposer.create
 */

/**
 * Creates an instance of a DigComposer with the specified
 * text.
 *
 * @param text (String) The HTML text to parse
 * @constructor
 * @private
 */
DigComposer = function(text) {
    this._tokens = new Object();
    this._tokenBlocks = new Object();
    this._chunks = new Array();
    // These attributes only apply to "mini me" composers
    this._parent = null;
    this._name = null;
    this._nameWithDelimeter = null;
    this._endNameWithDelimeter = null;
    if (text) {
        // Parse the text into a data structure that stores the
        // "chunks" of the html stream
        this._parse(text);
    }
}

// -----------------------------------------------------------
// Constants
// -----------------------------------------------------------

/**
 * Token engine parsing regex.  This regex pattern searches for
 * tokens and token blocks.  
 *
 * The original implementation looped over exec() calls on this regex 
 * using the "gm" options, global and multiline. 
 * Problems: IE 5.0 and IE/MAC parser fails on 'm' option
 * Problems: NS6 doesn't set the 'lastIndex' attribute on the returned
 *           result object, only on the Regex object itself.
 * Problems: IE 5.0 doesn't set the lastIndex attribute anywhere
 *
 * The new implementation loops() over exec() calls on the
 * following Regex manually adjusting the "text" passed to exec at
 * the end of each iteration and manually derives the "lastIndex."
 */
DigComposer._PARSER_REGEX = '@(\\w+)@|<!-- @\\/?(\\w+)@ -->';

// -----------------------------------------------------------
// API Methods
// -----------------------------------------------------------

/**
 * Returns an array containing the names of *all* tokens in
 * this composer including tokens in token blocks.
 *
 * @return (Array) The list of token names or an empty array if
 * the composer has no tokens
 */
DigComposer.prototype.getTokenNames = function() {
    return HTMLComposer.keys(this._tokens);
}

/**
 * Returns an array containing the names of *all* token blocks in
 * this composer including nested token blocks.
 *
 * @return (Array) The list of token block names or an empty array if
 * the composer has no token blocks
 */
DigComposer.prototype.getTokenBlockNames = function() {
    return HTMLComposer.keys(this._tokenBlocks);
}

/**
 * Returns true if the composer contains the specified token.
 *
 * @param name (String) name of token
 * @return (Boolean)
 */
DigComposer.prototype.containsToken = function(name) {

    // Optimized key lookup assuming keys never have undefined values
    return (typeof(this._tokens[name]) != 'undefined'); 
}

/**
 * Returns true if the composer contains the specified token block.
 *
 * @param name (String) name of token
 * @return (Boolean)
 */
DigComposer.prototype.containsTokenBlock = function(name) {

    // Optimized key lookup assuming keys never have undefined values
    return (typeof(this._tokenBlocks[name]) != 'undefined'); 
}

/**
 * Replaces all occurrences of the specified token with the passed
 * value. If value is null, the token will be removed from the
 * generated content.
 *
 * @param name (String) name of token
 * @param value (String) value of the token 
 * @param isOptionalToken (Boolean) If true and the token
 *        doesn't exist in this composer, this method will no-op
 *        (e.g. conveniance for calling 
 *        <code>HTMLComposer.containsToken()</code>
 *        before calling <code>HTMLComposer.replaceToken()</code>.  If
 *        false [default] and the token doesn't exist, a warning is
 *        logged to the composer category.
 */
DigComposer.prototype.replaceToken = function(name, value, 
                                              isOptionalToken) {
    // Optmized assertions!  These methods can be called 500+ times
    // Store the value in the map, used when its composed
    if (value === null) value = '';
    this._tokens[name] = value;
}

/**
 * Replaces all occurrences of the specified token block with the
 * passed value.  If value is null, the token block will be
 * removed from the generated content.
 *
 * @param name (String) name of token
 * @param value (String) value of the token 
 * @param isOptionalTokenBlock (Boolean) If true and the token block
 *        doesn't exist in this composer, this method will no-op
 *        (e.g. conveniance for calling 
 *        <code>HTMLComposer.containsTokenBlock()</code>
 *        before calling <code>HTMLComposer.replaceTokenBlock()</code>.  If
 *        false [default] and the token block doesn't exist, an
 *        error is logged to the composer category.
 */
DigComposer.prototype.replaceTokenBlock = function(
                                             name, value,
                                             isOptionalTokenBlock) {
    // Optmized assertions!  These methods can be called 500+ times
    // Store the value in the map, used when its composed
    if (value === null) value = '';
    this._tokenBlocks[name].value = value;
}

/**
 * <p>Returns a DigComposer representing the text within the
 * specified token block. The block returned by this method is
 * used to loop over its contents by calling <code>next()<code> on
 * the returned token block. See the documentation of this class
 * for an example.</p>
 *
 * @param name (String) name of token block
 * @return (DigComposer) 
 */
DigComposer.prototype.getTokenBlock = function(name) {

    // Return the token block reference
    return this._tokenBlocks[name].block;
}

/**
 * <p>Composes the contents of this token block into an internal
 * buffer.  This method is used to "loop" over the contents of a block.
 * The token block is eventually replaced with all of its
 * iterations once the main composer is composed.
 * See the documentation of this class for an example.</p>
 *
 * <p><i><b>Note:</b> This method can only be invoked on a token block,
 * that is a composer returned from
 * <code>getTokenBlock</code>.</i></p>
 */
DigComposer.prototype.next = function() {
    // Lazily create the array to store the iterations of this block
    var data = this._parent._tokenBlocks[this._name];
    if (!data.iterations) {
        data.iterations = new Array();
    }
    data.iterations[data.iterations.length] = this.compose();
}

/**
 * <p>Composes the HTML content of this composer, replacing all the
 * dynamic regions with their corresponding replaced values.</p>
 *
 * <p>Unreplaced tokens are removed from generated HTML.  The
 * contents of unreplaced token blocks are included in the
 * generated HTML.</p>
 *
 * <p>Once the generated content is returned as a string, the
 * composer resets itself.  That is, all previous replacements are
 * cleared.</p>
 *
 * @oparam anID (String) The contents of the element with this ID
 *         is set to the generated content.  Works for any element
 *         with a innerHTML property or for IFRAME/FRAME element.  If
 *         <code>anID</code> refers to an IFRAME element, the
 *         IFrame must have a "name" and an "id" attribute set in
 *         the HTML.
 * @return (String) The generated content.
 */
DigComposer.prototype.compose = function(anID) {
    var contents = this._compose(this);
    if (anID) {
        var elem = document.getElementById(anID);
        if ((elem.tagName == 'IFRAME') || HTMLComposer.isUndefined(elem.tagName)) {
            var aFrame = 
                (elem.tagName == 'IFRAME') 
                    ? window.frames[elem.id] : elem;
            aFrame.document.write(contents);
        } else {
            elem.innerHTML = contents;
        }
    }

    // Clear any dynamic values set by the caller to *reset* the
    // state of the HTMLComposer.  All other state is immutable.
    HTMLComposer.clearValues(this._tokens);
    for (var key in this._tokenBlocks) {
        this._tokenBlocks[key].value = null;
        this._tokenBlocks[key].iterations = null;
    }

    return contents;
}

// -----------------------------------------------------------
// Internal Methods
// -----------------------------------------------------------

/**
 * @param topComposer (DigComposer) The top composer where the
 *        replaced token and token block values are maintained
 * @return token (String) replaced string
 */
DigComposer.prototype._compose = function(topComposer) {
    var contents = new Array(this._chunks.length);
    for (var i = 0; i < this._chunks.length; i++) {
        var chunkValue = null;
        var chunk = this._chunks[i];
        if (HTMLComposer.isString(chunk)) {
            // Static Text 
            chunkValue = chunk;
        } else if (HTMLComposer.isInstance(chunk, _Token)) {
            // Token 
            var val = topComposer._tokens[chunk._name];
            if (val == null) {
                // Unreplaced token, remove the token from the
                // generated content and log a warning!
                val = '';
            }
            chunkValue = val;
        } else {
            // Token Block (e.g. DigComposer)

            // Get the replace value: 
            // First look to see if the token
            // block was explicitly replaced via replaceTokenBlock().
            // Then look to see if the token block has been
            // "looped" over via next().
            var blockData = topComposer._tokenBlocks[chunk._name];
            var val = blockData.value;
            if ((val == null) && blockData.iterations) {
                // We have iterations for the token block
                val = blockData.iterations.join('');
            }

            // If the block has been replaced, use the replaced value
            // The block wasn't replaced, compose its contents
            chunkValue = (val != null) ? val : chunk._compose(topComposer);
        }
        contents[i] = chunkValue;
    }

    return contents.join('');
}

/**
 * Clones this composer by clearing any replaced token and token
 * block values.
 */
DigComposer.prototype._clone = function() {
    var newComposer = new DigComposer();
    newComposer._chunks = this._chunks;
    // This clone is only invoked on token blocks and composers
    // coming from the cache.  Its therefore impossible for the
    // tokens and token block values to have values. 
    // Shallow reference is ok.
    newComposer._tokens = this._tokens;
    newComposer._tokenBlocks = this._tokenBlocks;
    return newComposer;
}

// -----------------------------------------------------------
// Parsing Methods
// -----------------------------------------------------------

/**
 * Parses the text and stores the "chunks" (text, token, token
 * block) in an array.
 */
DigComposer.prototype._parse = function(text) {
    var stack = new Array();
    var result = null;
    var index = 0;

    // Add this composer to the top of the composer stack.  The
    // top of the stack is the current HTMLComposer.  Required for
    // token blocks.
    stack.add(this);
    // Look for the static chunks, tokens, and token blocks
    var regex = new RegExp(DigComposer._PARSER_REGEX);
    while ((result = regex.exec(text)) != null) {
        var container = stack[stack.length - 1];
        // Add the static text before this token/token block
        container._addChunk(text.substring(0, result.index));
        if (!(result[0].indexOf("<!--") == 0)) {
            // Token
            container._addChunk(new _Token(result[1], result[0]));
        } else {
            // Token Block
            if (!(result[0].indexOf("<!-- @\/") == 0)) {
                // Start Token Block
                var block = new DigComposer();
                block._nameWithDelimeter = result[0];
                block._name = result[2];
                block._parent = container;
                container._addChunk(block);
                // Push this token block on the container stack
                stack.add(block);
            } else {
                // End Token Block
                container._endNameWithDelimeter = result[0];
                // Remove this token block from the container stack
                stack.length--;
            }
        } 
        // I would use result.lastIndex, but unfortunetly NS6
        // doesn't set this attribute on the result object
        // returned from the exec statement!  Manually compute the
        // last index
        text = text.substring(result.index + result[0].length);
    }

    // Add the static text after the last token/token block
    stack[stack.length - 1]._addChunk(text);
}

/**
 * Add the chunk into this composers immutable list of chunks.
 * If the chunk is a token or token block, add it to their
 * corresponding collections.
 */
DigComposer.prototype._addChunk = function(chunk) {
    this._addToCache(chunk);
    this._chunks.add(chunk);
}

/**
 * Adds token or token block chunks to the cache that maintains
 * the tokens and token blocks in this HTMLComposer.
 */
DigComposer.prototype._addToCache = function(chunk) {
    if (HTMLComposer.isInstance(chunk, _Token)) {
        this._tokens[chunk._name] = null;
    } else if (HTMLComposer.isInstance(chunk, DigComposer)) {
        // For token blocks, store the actual reference to the
        // block to allow for O(1) access to block references
        // needed for getTokenBlock().
        // Attributes for a token block:
        //    - block: the reference to the block
        //    - value: the replaced value (e.g. replaceTokenBlock())
        //    - stores the iterations of a token block (e.g. next())
        this._tokenBlocks[chunk._name] = 
            { value: null, block: chunk, iterations: null};
    }
    // Also add this chunk to the parents cache.  The parent
    // composer always maintain a flat list of tokens and token blocks.
    if (this._parent) {
        this._parent._addToCache(chunk);
    }
}




/**
 * Array extensions are additions to the built-in JavaScript Array object.
 *
 * @class Array
 */

// -----------------------------------------
// Extension Methods
// -----------------------------------------

/**
 * Add <code>anObject</code> to the end of an array.
 * 
 * @param anObject The object to add to the end of the array.
 */
Array.prototype.add = function(anObject) {
    this[this.length] = anObject;
}

/**
 * <p>Add <code>anObject</code> to the end of an array only if it
 * is not already contained in the array.</p>
 * 
 * @param anObject The object to add to the end of the array
 */
Array.prototype.addUnique = function(anObject) {
    if (!this.contains(anObject)) {
        this[this.length] = anObject;
    }
}

/**
 * Add all the objects in <code>anArray</code> to the end of this
 * array.
 *
 * @param anArray (Array) The objects to add.
 */
Array.prototype.addAll = function(anArray) {
    for (var i = 0; i < anArray.length; i++) {
        this.add(anArray[i]);
    }
}

/**
 * Insert a specified object at a specified index in the
 * array.  Shifts the element currently at that position and any
 * subsequent elements to the right.
 *
 * @param anObject the object to insert into the array.
 * @param anIndex (Number) the index where the object will be
 *                inserted.  The index must be >= 0 and <= length of 
 *                the array.
 */
Array.prototype.insert = function(anObject, anIndex) {
    // Shift any existing objects
    for (var i = this.length - 1; i >= anIndex; i--) {
            this[i+1] = this[i];
    }
    // Insert the object passed by the caller
    this[anIndex] = anObject;
}

/**
 * Remove the first occurrence of <code>anObject</code> from the array.
 * 
 * @param anObject The object to remove.
 * @return the object if it is successfully removed, null otherwise
 *         (i.e. no matching object found).
 */
Array.prototype.remove = function(anObject) {
    var retval = null;
    var myIndex = this.indexOf(anObject);
    if (myIndex != -1) {
        this.removeAt(this.indexOf(anObject));
        retval = anObject;
    } 

    return retval;
}

/**
 * Remove the object at <code>anIndex</code> and shift trailing objects
 * to take up the vacated slot.
 *
 * @param anIndex (Number) remove the object at this index.
 * @return the object that is removed, null otherwise.
 */
Array.prototype.removeAt = function(anIndex) {
    var myObject = this[anIndex];
    for (var i = anIndex; i < this.length-1; i++) {
        this[i] = this[i+1];
    }

    delete this[(this.length-1)];
    this.length--;

    return myObject;
}

/**
 * Remove the last object in the array.
 * 
 * @return (Object) the object removed from the array, null if no objects 
 *         in the array.
 */
Array.prototype.removeLast = function() {
    if (this.length <= 0) return null;
    var myObject = this[this.length - 1];
    this.removeAt(this.length - 1);

    return myObject;
}

/**
 * Remove all the objects in <code>anArray</code> from this array.
 *
 * @param anArray (Array) The objects to remove.
 */
Array.prototype.removeAll = function(anArray) {
    for (var i = 0; i < anArray.length; i++) {
        this.remove(anArray[i]);
    }
}

/**
 * Clear the array by removing *all* the elements.
 */
Array.prototype.clear = function() {
    for (var i = this.length-1; i >= 0; i--) {
    delete this[i];
    }
    this.length = 0;
}

/**
 * Return the last object in the array or null if it's empty.
 * @return 
 */
Array.prototype.last = function() {
    if (this.length <= 0) return null;
    return this[this.length - 1];
}

/**
 * Return the index of <code>anObject</code> if it's in the array,
 * otherwise return -1.
 *
 * @param anObject the object to find in this array
 * @return (Number) the index of the object or -1 if it's not found.
 */
Array.prototype.indexOf = function(anObject) {
    for (var i = 0; i < this.length; i++) {
        if (this[i] == anObject) return i;
    }

    return -1;
}

/**
 * Return true if <code>anObject</code> is contained in the array.
 *
 * @param anObject the object to find in this array.
 * @return (Boolean) true if the object is contained in the array.
 */
Array.prototype.contains = function(anObject) {
    return (this.indexOf(anObject) != -1);
}

/**
 * Return a shallow copy of this array.
 *
 * @return (Array) a shallow copy of this array.
 */
Array.prototype.clone = function() {
    var newArray = new Array(this.length);
    for (var i = 0; i < this.length; i++) newArray[i] = this[i];

    return newArray;
}

/**
 * <p>Default comparator to use in searching if the caller did not
 * provide one.  Compares <code>a</code> and <code>b</code> using
 * the built in JavaScript operators (e.g. "<", ">").</p>
 *
 * @param a key being searched for
 * @param b key to compare to 
 * @see Array.binarySearch
 */
Array.defaultComparator = function(a, b) {
    return (a < b) ? -1 :((a > b) ? 1 : 0);
}

