/* -*- tab-width: 4 -*- make emacs play nice  */
//#
//# Copyright (c) 2001 AOL Corporation. All Rights Reserved.
//#
/**
 * SIButton provides an interface to
 * generically handle the "enablement" state of a given button. 
 *
 * This class has been tested on IE 5.0+, NS 6.1+.
 *
 * @class SIButton
 */
// ----------------------------------------------------
//  Creation
// ----------------------------------------------------
/**
 * @constructor
 */



function SIButton() {};

/**
 * Factory method to create a button. The button's images are derived
 * from its corresponding image element in the this.getDoc(). The img tag
 * can have "enabledsrc" and "disabledsrc" attributes which the SIButton
 * will use for its enabled and disabled states, respectively. If either
 * of these is missing, the "src" attribute will be used in its place.
 * This allows you to specify the images for a SIButton in the HTML and
 * allows you to set the initial visible state of the button to either
 * enabled or disabled (e.g. to make the button appear disabled initially,
 * set the img tag's src to the disabled image and its enabledsrc to the
 * enabled image).
 *
 * @param id 
 *     id of the button (e.g. "read").  The DOM should have an
 *     image for this button.  The id of the image is derived 
 *     from the 'name' (e.g. "readImage").
 * @oparam onclickHandler  
 *     function to invoke when the user clicks on the button
 *     (e.g. 'readSelectedMessages()')
 *
 * @see updateEnablement()
 * @return SIButton
 */
SIButton.map = new Object();

SIButton.kNormal = 0;
SIButton.kClicked = 1;
SIButton.kDisabled = 2;
SIButton.kHover = 3;

SIButton.kTypeNormal = 0;
SIButton.kTypeToggle = 1;
SIButton.kTypeSticky = 2;

SIButton.onContextMenu = function() {
    return false;
}
document.oncontextmenu = SIButton.onContextMenu;

SIButton.altTextMap = {
"openIOCP":SIConfig.labInternet,
"openTM":SIConfig.labTask,
"openAR":SIConfig.labAdd ,
"openDM":SIConfig.labDevice, 
"openSM":SIConfig.labServices, 
"openDP":SIConfig.labColor, 
"openWU":SIConfig.labWebUtil, 
"openNC":SIConfig.labNetwork, 
"openCU":SIConfig.labCheckup,
"startRenew":SIConfig.labRelease,
"clearCache":SIConfig.labCurrent
}


SIButton.createClipped = function(id, onclickHandler, type, state, doc) {
    var button = new SIButton();
    button.width = null;
    button.height = null;
    
    button.doc = (doc) ? doc : null;
    button.type = (type) ? type : SIButton.kTypeNormal;
    button.state = (state) ? state : SIButton.kNormal;

    var created = button._init(id, onclickHandler);
    button.enable();
    button._setCell(button.state);
    return button;
}

SIButton.isSpecified = function(aThing) {
    return ((aThing !== void 0) && (aThing !== null));
}

/**
 * Initializes the state of this button.
 *
 * Instance Variables:
 *    _id                 - the DOM id for this button (e.g. "read")
 *                          The actual ID of the image element is
 *                          derived from this _id by appending
 *                          'Image' 
 *    _shouldEnableFunc   - the function that implements the enablement logifc
 * @return boolean whether the button was created
 */
SIButton.prototype.getDoc = function() {
    if (this.doc) {
        return this.doc;
    }
    else {
        return document;
    }
}

SIButton.prototype._handleLink = function() {
    if (event.keyCode == 13 || event.type == 'click') {
        var button = SIButton.map[this.id.substr(1)];
        if (SIButton.isSpecified(button)) {
            if (button.isEnabled()) {
                button.onclickHandler(button._id);
            } 
        }
    }
    return  false;
}

SIButton.prototype._init = function(id, onclickHandler) {
    
    this._id = id;
    this._exists = true;
    var image = this._getImageElement();
    if (!SIButton.isSpecified(image)) {
        // don't hose everything by asserting
        this._exists = false;
    } else {
        SIButton.map[image.id] = this;
        if (SIButton.isSpecified(image.style.top) && (image.style.top != "")){
            this.origTop = parseInt(image.style.top);
        } else {
            this.origTop = 0;
        }
        this._shouldEnableFunc = null;
        this._mouseDownNotify = null;
                
        var rect = image.style.clip.toLowerCase();
        var aStr = rect.replace(/[, ]+/g,',').replace(/px/g,'').replace(/rect/,'').replace(/[\(\)]/g,'');
        var dimen = aStr.split(',');
        this.height = parseInt(dimen[2]);
        this.width = parseInt(dimen[1]);
        var path = image.src.split('/');
        var file = path[path.length-1];
        var fileroot = file.substr(0, file.toLowerCase().indexOf('.gif') );
        var fileroot = fileroot.replace(/_/g, '');
        var fileroot = fileroot.replace(/-/g, '');
        
        var bLabel = SIConfig.buttonLabels[fileroot];
        if (SIButton.altTextMap[id]) {
            bLabel += ' ' + SIButton.altTextMap[id];
        }
        this.onclickHandler = onclickHandler;
        image.onmousedown = SIButton._mousedown;
        image.onmouseup = SIButton._mouseup;
        image.onmouseover = SIButton._mouseover;
        image.onmouseout = SIButton._mouseout;
        image.alt = bLabel;
        var aTag = this.getDoc().createElement('a');
        aTag.title = bLabel;
        aTag.id = '_' +  id;
        aTag.target = '_self';
        aTag.href = 'javascript:void';
        aTag.onkeypress = this._handleLink; 
        aTag.onclick = this._handleLink; 
        aTag.style.position = 'absolute';
        aTag.style.zIndex = image.style.zIndex + 5;
        aTag.style.top = parseInt(image.style.top) + 1 + 'px';
        aTag.style.left = parseInt(image.style.left) + 1 + 'px';
        aTag.style.height = this.height - 2 + 'px';
        aTag.style.width = this.width - 2 + 'px';
        aTag.style.lineHeight = '0px';        
        image.insertAdjacentElement('afterEnd', aTag);   
    }
    return true;
}


// ----------------------------------------------------
//  API Methods
// ----------------------------------------------------


/**
 * Return the id of this button.
 */
SIButton.prototype.getID = function() {
    return this._id;
}

/**
 *  shouldEnableFunc 
 *     the function that determines the enablement state.  If a 
 *     function ptr isn't provided, the button will always be enabled.
 * 
 */
SIButton.prototype.setShouldEnableFunction = function(handler) {
    this._shouldEnableFunc = handler;
}

/**
 * Function to call when the mouse is clicked down
 * Note: onclickHandler is called on mouse up.
 */
SIButton.prototype.setMouseDownNotify = function(handler) {
    this._mouseDownNotify = handler;
}

/**
 * Update the enablement state of this button
 */
SIButton.prototype.updateEnablement = function() {
    if (this._exists) {
        if (this._shouldEnable()) {
            this.enable();
        } else {
            this.disable();
        }
    }
}

/**
 * Enable this button
 */
SIButton.prototype.enable = function() {
    // This allows the user to click on the button and 
    // gives a UI indication of this state
    this._currentlyEnabled = true;
    if (this._exists) {
        this._setCell(SIButton.kNormal);
    }
}

/**
 * Disable this button
 */
SIButton.prototype.disable = function() {
    // Display the disabled image
    this._currentlyEnabled = false;
    if (this._exists) {
        this._setCell(SIButton.kDisabled);
    }
}

/**
 * Enable or disable this button
 *
 * @param enable (Boolean)
 */
SIButton.prototype.setEnabled = function(enable) {
    if (this._exists) {
        if (enable) {
            this.enable();
        } else {
            this.disable();
        }
    }
}

/**
 * Is this button enabled?
 */
SIButton.prototype.isEnabled = function () {
    return (this._exists && this._currentlyEnabled);
}

SIButton.prototype.toString = function() {
    return "SIButton " + this._id;
}

// ----------------------------------------------------
//  Internal Methods
// ----------------------------------------------------

SIButton.prototype._setCell = function(cellNum) {
    // if the button doesn't really exist, return immediately
    if (!this._exists) {
        return;
    }
    if (!this.isEnabled()) cellNum = SIButton.kDisabled;
    var img = this._getImageElement();
            
    var clipTop =  cellNum*this.height;
    var clipRight = this.width;
    var clipBottom = clipTop + this.height;
    var clipLeft = 0;
    
    img.style.clip = "rect(" + clipTop + "px " + clipRight + "px " + clipBottom + "px " + clipLeft + "px " + ")";
    img.style.top =  -clipTop + this.origTop + "px";
}

/**
 * @return true if this button should be enabled
 */
SIButton.prototype._shouldEnable = function() {
    // If the caller didn't provide an enablement callback, always
    // have the button enabled!
    // Otherwise delegate to the func passed in at creation time
    var retval = true;
    if (this._exists) {
        if (this._shouldEnableFunc) {
            retval = this._shouldEnableFunc();
        }
    } else {
        retval = false;
    }
    return retval;
}

SIButton.prototype.setButtonState = function(state) {
    if (this._exists) {
        this.state = state;
        if (state) {
            this._setCell(SIButton.kClicked);
        }
        else {
            this._setCell(SIButton.kNormal);
        }
    } else {
        return void 0;
    }
}

SIButton.prototype.getButtonState = function() {
    if (this._exists) {
        return this.state;
    } else {
        return void 0;
    }
}

/**
 * @return this button's image element
 */
SIButton.prototype._getImageElement = function() {
    if (this._exists) {
        return this.getDoc().getElementById(this._id);
    } else {
        return void 0;
    }
}

SIButton._mousedown = function() {
    var button = SIButton.map[this.id];
    if (SIButton.isSpecified(button)) {
        if (button.isEnabled()) {
            button._setCell(SIButton.kClicked);
            if (button._mouseDownNotify) {
                button._mouseDownNotify(button._id);
            }
        }
    }
}

SIButton._mouseup = function() {
    var button = SIButton.map[this.id];
    if (SIButton.isSpecified(button)) {
        if (button.isEnabled()) {
            if (button.type == SIButton.kTypeSticky) {
                button.setButtonState(true);
            }
            else if (button.type == SIButton.kTypeToggle) {
                button.setButtonState( !button.getButtonState() );                
            }
            else {
                button._setCell(SIButton.kNormal);
            }
            button.onclickHandler(button._id);
        } 
    }
}

SIButton._mouseover = function() {
    var button = SIButton.map[this.id];
    if (SIButton.isSpecified(button)) {
        if (button.isEnabled()) {
            if (button.type == SIButton.kTypeNormal || !button.getButtonState())
                button._setCell(SIButton.kHover);
        }
    }
}

SIButton._mouseout = function() {
    var button = SIButton.map[this.id];
    if (SIButton.isSpecified(button)) {
        if (button.isEnabled()) {
            if (button.type == SIButton.kTypeNormal || !button.getButtonState())
                button._setCell(SIButton.kNormal);
        }
    }
}
